"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const cloudformation_1 = __importDefault(require("aws-sdk/clients/cloudformation"));
const through2_1 = __importDefault(require("through2"));
const plugin_error_1 = __importDefault(require("plugin-error"));
const fancy_log_1 = __importDefault(require("fancy-log"));
const ansi_colors_1 = __importDefault(require("ansi-colors"));
const fromPairs_1 = __importDefault(require("lodash/fromPairs"));
const takeWhile_1 = __importDefault(require("lodash/takeWhile"));
const startsWith_1 = __importDefault(require("lodash/startsWith"));
const endsWith_1 = __importDefault(require("lodash/endsWith"));
const includes_1 = __importDefault(require("lodash/includes"));
const MAX_REPORT_SECONDS = 15;
const error = (err) => new plugin_error_1.default('cf-deploy', err, { showProperties: typeof err !== 'string' });
const isInProgress = (state) => endsWith_1.default(state.StackStatus, '_IN_PROGRESS');
const isCleanupInProgress = (state) => endsWith_1.default(state.StackStatus, '_CLEANUP_IN_PROGRESS');
const isInitialStackEvent = (event, StackId) => event.PhysicalResourceId === StackId &&
    ['CREATE_IN_PROGRESS', 'UPDATE_IN_PROGRESS'].some(s => s === event.ResourceStatus);
const isFailedStackEvent = (event) => endsWith_1.default(event.ResourceStatus, 'FAILED');
const stackStateIndicatesFailure = (state) => ['ROLLBACK', 'DELETE', 'FAILED'].some(s => includes_1.default(state.StackStatus, s));
const isDeploySuccessful = (state) => includes_1.default(state.StackStatus, 'COMPLETE') &&
    !includes_1.default(state.StackStatus, 'ROLLBACK') &&
    ['CREATE', 'UPDATE'].some(s => startsWith_1.default(state.StackStatus, s));
const simplifiedOutput = (state) => fromPairs_1.default(state.Outputs.map(({ OutputKey, OutputValue }) => [
    OutputKey,
    OutputValue,
]));
const parameterValue = (param) => String(param);
// noinspection JSUnusedGlobalSymbols
exports.default = (serviceOptions, stackNameOrOptions, parameters = {}, options = {}) => through2_1.default.obj(async (file, enc, done) => {
    const deploy = async () => {
        const stackOptions = typeof stackNameOrOptions === 'string'
            ? { StackName: stackNameOrOptions }
            : typeof stackNameOrOptions === 'undefined'
                ? { StackName: file.stem }
                : stackNameOrOptions;
        if (file.isNull()) {
            return file;
        }
        if (!file.isBuffer()) {
            throw error('Can only handle buffered files, pipe through vinyl-buffer beforehand');
        }
        const stackName = stackOptions.StackName;
        const logPrefix = options.logPrefix || stackName;
        const cfn = new cloudformation_1.default(Object.assign({ apiVersion: '2010-05-15' }, serviceOptions));
        const retrieveStackEvents = async (stackId) => {
            const eventsResult = await cfn
                .describeStackEvents({ StackName: stackId })
                .promise();
            return takeWhile_1.default(eventsResult.StackEvents, e => !isInitialStackEvent(e, stackId)).reverse(); // Change to chronological order
        };
        const logFailures = (stackEvents) => {
            stackEvents
                .filter(e => isFailedStackEvent(e))
                .forEach(e => fancy_log_1.default(`${logPrefix}.${e.LogicalResourceId}:`, ansi_colors_1.default.dim(`(${e.ResourceStatus})`), ansi_colors_1.default.red(e.ResourceStatusReason || e.ResourceStatus || '<unknown>')));
        };
        const buildParameters = () => [
            ...(stackOptions.Parameters || []),
            ...Object.entries(parameters).map(([k, v]) => ({
                ParameterKey: k,
                ParameterValue: parameterValue(v),
            })),
        ];
        const completedState = async ({ StackId, reportEvents, waitForCleanup, }) => {
            let haveReportedFailures = false;
            const reportedLogicalIds = [];
            let state;
            let delaySeconds = 2;
            for (;;) {
                const describeResult = await cfn
                    .describeStacks({ StackName: StackId })
                    .promise();
                state = describeResult.Stacks[0];
                const completed = !isInProgress(state) ||
                    (!waitForCleanup && isCleanupInProgress(state));
                if (reportEvents) {
                    const stackEvents = await retrieveStackEvents(StackId);
                    stackEvents
                        .filter(e => e.LogicalResourceId && e.PhysicalResourceId !== StackId)
                        .forEach(e => {
                        if (!includes_1.default(reportedLogicalIds, e.LogicalResourceId)) {
                            const prefix = `${logPrefix}.${e.LogicalResourceId}`;
                            if (startsWith_1.default(e.ResourceStatus, 'CREATE')) {
                                fancy_log_1.default(`${prefix}: ${ansi_colors_1.default.green('✔ creating')}`);
                            }
                            else if (startsWith_1.default(e.ResourceStatus, 'UPDATE')) {
                                fancy_log_1.default(`${prefix}: ✱ updating`);
                            }
                            else if (startsWith_1.default(e.ResourceStatus, 'DELETE')) {
                                fancy_log_1.default(`${prefix}: ${ansi_colors_1.default.dim('✘ deleting')}`);
                            }
                            reportedLogicalIds.push(e.LogicalResourceId);
                        }
                    });
                    if (!haveReportedFailures && stackStateIndicatesFailure(state)) {
                        logFailures(stackEvents);
                        fancy_log_1.default('See full details at:', ansi_colors_1.default.blue.underline(`https://${serviceOptions.region}.console.aws.amazon.com/cloudformation/home?region=${serviceOptions.region}#/stack/detail?stackId=${encodeURIComponent(StackId)}`));
                        haveReportedFailures = true;
                    }
                }
                fancy_log_1.default(`${logPrefix}: ${state.StackStatus}`, completed
                    ? ''
                    : ansi_colors_1.default.dim(`checking again in ${delaySeconds}s...`));
                if (completed) {
                    return state;
                }
                await new Promise(resolve => setTimeout(resolve, delaySeconds * 1000)); // eslint-disable-line
                delaySeconds = Math.min(delaySeconds * 2, MAX_REPORT_SECONDS);
            }
        };
        let initialState;
        try {
            const describeResult = await cfn
                .describeStacks({ StackName: stackName })
                .promise();
            initialState = describeResult.Stacks[0];
        }
        catch (e) {
            if (e.code !== 'ValidationError') {
                throw e;
            }
        }
        if (initialState) {
            if (isInProgress(initialState)) {
                fancy_log_1.default(`${logPrefix}: waiting for previous operation to complete`);
                initialState = await completedState({
                    StackId: initialState.StackId,
                    reportEvents: false,
                    waitForCleanup: true,
                });
                fancy_log_1.default(`${logPrefix}: starting deployment`);
            }
        }
        else {
            fancy_log_1.default(`${logPrefix}: creating new stack`);
        }
        const updating = initialState != null;
        const deployParams = Object.assign({}, (updating ? {} : { OnFailure: 'DELETE' }), stackOptions, { Parameters: buildParameters(), TemplateBody: file.contents.toString(enc) });
        let resultState;
        let skipped = false;
        try {
            const request = updating
                ? cfn.updateStack(deployParams)
                : cfn.createStack(deployParams);
            const deployResult = await request.promise();
            resultState = await completedState({
                StackId: deployResult.StackId,
                reportEvents: true,
                waitForCleanup: false,
            });
        }
        catch (e) {
            if (updating &&
                e.code === 'ValidationError' &&
                e.message === 'No updates are to be performed.') {
                // CloudFormation will only update if *resources* will change, for example: changes to Outputs don't count
                resultState = initialState; // Must exist since we're updating
                skipped = true;
            }
            else {
                throw e;
            }
        }
        if (!skipped && !isDeploySuccessful(resultState)) {
            if (endsWith_1.default(resultState.StackStatus, '_FAILED')) {
                logFailures(await retrieveStackEvents(resultState.StackId));
            }
            throw error(`Deploying ${stackName} failed (${resultState.StackStatusReason ||
                resultState.StackStatus})`);
        }
        const output = simplifiedOutput(resultState);
        const outputFile = file.clone({ contents: false });
        outputFile.contents = Buffer.from(JSON.stringify(output, null, 2), 'utf8');
        outputFile.data = output;
        outputFile.extname = 'json';
        return outputFile;
    };
    try {
        done(null, await deploy());
    }
    catch (e) {
        if (e instanceof plugin_error_1.default) {
            done(e);
        }
        else {
            done(error(e));
        }
    }
});
